package ch.epfl.lamp.minischeme;

import java.io._;
import java.text.DecimalFormat;
import scala.collection.mutable._;

/**
 * Minivm code sequence.
 *
 * @author Michel Schinz
 */
class Code {
  abstract class Item { def printTo(stream: PrintStream): Unit; }
  case class InstructionItem(instruction: Instruction) extends Item {
    override def printTo(stream: PrintStream): Unit =
      instruction printTo stream;
  }
  case class LabelItem(label: Label) extends Item {
    override def printTo(stream: PrintStream): Unit =
      label printTo stream;
  }
  case class CommentItem(comment: String) extends Item {
    override def printTo(stream: PrintStream): Unit =
      stream.println("; " + comment);
  }

  val WORD_SIZE = 4;

  // Instructions
  private val items = new ListBuffer[Item]();

  def emit(i: Instruction): Unit =
    items += InstructionItem(i);
  def emit(op: OpcodeRRR, r1: Register, r2: Register, r3: Register): Unit =
    emit(new InstructionRRR(op, r1, r2, r3));
  def emit(op: OpcodeRRC, r1: Register, r2: Register, l: Label): Unit =
    emit(new InstructionRRC(op, r1, r2, LabelConstant(l)));
  def emit(op: OpcodeRRC, r1: Register, r2: Register, i: Int): Unit =
    emit(new InstructionRRC(op, r1, r2, IntegerConstant(i)));
  def emit(op: ALOC.type, r1: Register, r2: Register): Unit =
    emit(new InstructionRR(op, r1, r2));
  def emit(op: LINT.type, r: Register, i: Int): Unit =
    emit(new InstructionRC(op, r, IntegerConstant(i)));
  def emit(op: LINT.type, r: Register, l: Label): Unit =
    emit(new InstructionRC(op, r, LabelConstant(l)));
  def emit(op: OpcodeR, r: Register): Unit =
    emit(new InstructionR(op, r));
  def emit(op: HALT.type): Unit =
    emit(new InstructionHalt(op));
  def emit(label: Label): Unit =
    items += LabelItem(label);

  def addComment(comment: String): Unit =
    items += CommentItem(comment);

  def emitPush(reg: Register): Unit = {
    emit(STOR, reg, FP, getFrameSize);
    incFrameSize(WORD_SIZE);
  }
  def emitPop(reg: Register): Unit = {
    decFrameSize(WORD_SIZE);
    emit(LOAD, reg, FP, getFrameSize);
  }
  def emitJump(targetLabel: Label, condReg: Register): Unit =
    withFreshRegister { tmpReg =>
      emit(LINT, tmpReg, targetLabel);
      emit(CMOV, PC, tmpReg, condReg)
    };

  // Registers
  val registers =
    for (val r <- List.range(0,32)) yield
      new Register(r);
  val R0 = registers(0);
  val LK = registers(28);
  val FP = registers(29);
  val GP = registers(30);
  val PC = registers(31);
  private val usableRegisters =
    for (val r <- List.range(1,28)) yield
      registers(r);

  private var freeRegisters: List[Register] = usableRegisters;

  def allocateRegister: Register = freeRegisters match {
    case r :: rest => freeRegisters = rest; r
    case _ => error("No more free registers")
  }
  def allocateRegister(r: Register): Register =
    if (freeRegisters contains r) {
      freeRegisters = freeRegisters diff List(r);
      r
    } else
      error("Register " + r + " already allocated");

  def allocatedRegisters: List[Register] =
    usableRegisters diff freeRegisters;

  def freeRegister(r: Register) =
    freeRegisters = r :: freeRegisters;

  def withFreshRegister[T](body: Register=>T): T = {
    val r = allocateRegister;
    try {
      body(r);
    } finally {
      freeRegister(r)
    }
  }

  // Labels
  private val labelCounters = new HashMap[String,Int];
  def newLabel: Label = newLabel("L");
  def newLabel(prefix: String): Label = {
    val sanePrefix = prefix.replace(':','_');
    if (!(labelCounters contains sanePrefix)) {
      labelCounters += sanePrefix -> 1;
      new Label(sanePrefix)
    } else {
      val count = labelCounters(sanePrefix);
      labelCounters += sanePrefix -> (count + 1);
      new Label(sanePrefix + count);
    }
  };

  // Frame size
  private var frameSize: Int = 0;
  def getFrameSize: Int = frameSize;
  private var maxFrameSize: Int = 0;
  def incFrameSize(delta: Int): Unit = {
    frameSize = frameSize + delta;
    if (frameSize > maxFrameSize)
      maxFrameSize = frameSize;
  }
  def decFrameSize(delta: Int): Unit =
    incFrameSize(-delta);
  def getMaxFrameSize: Int = maxFrameSize;
  def resetMaxFrameSize: Unit = (maxFrameSize = 0);

  // Dumping
  def dumpTo(stream: PrintStream): Unit = {
    for (val i <- items)
      i printTo stream;
  }
}
