package ch.epfl.lamp.minischeme;

import scala.collection.immutable.Map;

/**
 * Name analysis for minischeme.
 *
 * @author Michel Schinz
 */
object NameAnalyzer extends Function2[Map[String,Symbol],List[Tree],Unit] {
  private type Env = Map[String,Symbol];

  def apply(env: Env, trees: List[Tree]): Unit = {
    val globalEnv = collectGlobalEnv(env, trees);
    for (val tree <- trees) analyze(globalEnv, tree)
  }

  private def collectGlobalEnv(init: Env, trees: List[Tree]): Env =
    trees.foldLeft(init) { (env, tree) => tree match {
      case SList(lst) => lst match {
        case SIdent("define") :: (nameTree @ SIdent(name)) :: _ =>
          env + name -> new Symbol(name, GlobalKind)
        case _ => env
      }
      case _ =>
        env
    } }

  private def analyze(env: Env, tree: Tree): Unit = tree match {
    case SList(SIdent("define") :: SIdent(name) :: value :: Nil) =>
      val nameTree = tree.asInstanceOf[SList].elements(1);
      nameTree.symbol = env(name);
      analyze(env, value)

    case SList(SIdent("lambda") :: SList(parameters) :: body) =>
      val symbols = for (val parameter @ SIdent(name) <- parameters) yield {
        parameter.symbol = new Symbol(name, LocalKind);
        Pair(name, parameter.symbol)
      };
      checkDefs(tree.line, tree.column, symbols map (._1));
      val augmentedEnv = env incl symbols;
      for (val statement <- body)
        analyze(augmentedEnv, statement)

    case SList(SIdent("let") :: SList(bindings) :: body) =>
      val symbols = for (val SList(ident :: rhs :: Nil) <- bindings) yield {
        val SIdent(name) = ident;
        ident.symbol = new Symbol(name, LocalKind);
        analyze(env, rhs);
        Pair(name, ident.symbol)
      };
      checkDefs(tree.line, tree.column, symbols map (._1));
      val augmentedEnv = env incl symbols;
      for (val statement <- body)
        analyze(augmentedEnv, statement)

    case SList(SIdent("if") :: cond :: thenp :: elsep :: Nil) =>
      analyze(env, cond);
      analyze(env, thenp);
      analyze(env, elsep)

    case SList(elements) =>
      for (val element <- elements)
        analyze(env, element);

    case SIdent(name) =>
      if (env contains name)
        tree.symbol = env(name);
      else
        Report.error(tree.line, tree.column, "unknown identifier " + name);

    case SNum(_) =>
  }

  private def checkDefs(l: Int, c: Int, names: List[String]): Unit = {
    val defined = new scala.collection.mutable.HashSet[String]();
    for (val name <- names) {
      if (defined(name))
        Report.error(l, c, "multiple definition of " + name);
      defined += name;
    }
  }
}
