/*
 * MiniVM project (c) LAMP 2006
 * 
 * Execution engine.
 * 
 * $Id: vm.c,v 1.1 2006/03/16 15:55:26 dragos Exp $
 */

#include <stdlib.h>
#include "../config.h"
#include "../main.h"
#include "../log.h"
#include "../loader/loader.h"
#include "../loader/alignloader.h"
#include "../memory/memory.h"
#include "../util/pointers.h"

#include "vm.h"
#include "prompt.h"

/* Block of memory which contains the decoded program. */
char* code_mem = NULL;

/* Block of memory used for heap allocation. */
char* heap_mem = NULL;

/* The garbage collector */
garbage_collector_t* gc = NULL;

/* The address of the last instruction of the program. */
char* end_program = 0;

/* The registers. */
int R[REGISTERS];

/* The current opcode. */
static int opcode;

void check_heap_pointer(int addr);
void exec();
void dump_code();
void exit_vm();
void next_pc(int);


/*
 * Initialize the virtual machine: allocate memory and
 * call the loader module.
 */
void
init_vm()
{
  code_mem = malloc(code_size);
  if (code_mem == NULL)
    {
      printf("Could not allocate %dk of memory.\n", code_size / 1024);
      exit_vm();
    }
        
  write_log(LOG_DEBUG, "Allocated %4dk of memory for code at: 0x%0.8x\n", code_size / 1024, code_mem);
        
  heap_mem = malloc(heap_size);
  if (heap_mem == NULL)
    {
      printf("Could not allocate %dk of memory.\n", heap_size / 1024);
      exit_vm();
    }

  write_log(LOG_DEBUG, "Allocated %4dk of memory for heap at: 0x%0.8x up to: 0x%0.8x\n", heap_size / 1024, heap_mem, heap_mem + heap_size);
  write_log(LOG_INFO, "Parsing file %s\n", filename);
  
  program();
  write_log(LOG_INFO, "Program loaded successfully, beginning execution.\n");

  /* Initialize the gc. */
  if (gc == NULL)
    gc = &no_gc;
  gc->init();

    exec();
}

/* The execution loop. */
void
exec()
{
  /* initialize R31 to be PC */
  R[31] = (int)code_mem;
  R[0]  = 0;
  
  int a, b, c;
        
  /* the interpreter loop. */
  while (TRUE)
    {
      check_code_pointer(R[31]);
      if (interactive)
        prompt_user();
      opcode = decode(R[31], &a, &b, &c);
      next_pc(opcode);
      switch (opcode)
        {
        case OP_LINT:
          if (is_code_pointer(c))
            write_log(LOG_DEBUG, "\tLINT\tR%d\t0x%0.8x\n", a, c);
          else
            write_log(LOG_DEBUG, "\tLINT\tR%d\t%d\n", a, c);
          R[a] = c;
          break;
        case OP_ADD:
          write_log(LOG_DEBUG, "\tADD\tR%d\tR%d\tR%d\n", a, b, c);
          R[a] = R[b] + R[c];
          break;
        case OP_SUB:
          write_log(LOG_DEBUG, "\tSUB\tR%d\tR%d\tR%d\n", a, b, c);
          R[a] = R[b] - R[c];
          break;
        case OP_MUL:
          write_log(LOG_DEBUG, "\tMUL\tR%d\tR%d\tR%d\n", a, b, c);
          R[a] = R[b] * R[c];                             
          break;
        case OP_DIV:
          write_log(LOG_DEBUG, "\tDIV\tR%d\tR%d\tR%d\n", a, b, c);
          if (R[c] == 0)
            {
              printf("Division by zero at 0x%.8x\n", (int)PC);
              exit_vm();
            }
          R[a] = R[b] / R[c];                             
          break;
        case OP_MOD:
          write_log(LOG_DEBUG, "\tMOD\tR%d\tR%d\tR%d\n", a, b, c);
          if (R[c] == 0)
            {
              printf("Division by zero at 0x%.8x\n", (int)PC);
              exit_vm();
            }
          R[a] = R[b] % R[c];                             
          break;
                                
        case OP_ISLT:
          write_log(LOG_DEBUG, "\tISLT\tR%d\tR%d\tR%d\n", a, b, c);
          R[a] = R[b] < R[c];
          break;
        case OP_ISLE:
          write_log(LOG_DEBUG, "\tISLE\tR%d\tR%d\tR%d\n", a, b, c);
          R[a] = R[b] <= R[c];
          break;
        case OP_ISEQ:
          write_log(LOG_DEBUG, "\tISEQ\tR%d\tR%d\tR%d\n", a, b, c);
          R[a] = R[b] == R[c];
          break;
        // ...  complter ...
        
        case OP_ALOC:
          write_log(LOG_DEBUG, "\tALOC\tR%d\tR%d\n", a, b);
          R[a] = (int)gc->alloc(R[b]);
          break;
        case OP_LOAD:
          if (is_code_pointer(c))
            {
              write_log(LOG_DEBUG, "\tLOAD\tR%d\tR%d\t0x%0.8x\n", a, b, c);
              printf("Memory load from outside of heap segment at PC: 0x%.8x", (int)PC);
              exit_vm();
            } else
              {
                write_log(LOG_DEBUG, "\tLOAD\tR%d\tR%d\t%d\n", a, b, c);
                check_heap_pointer(R[b] + c);
                R[a] = *(int*)(R[b] + c);
              }
          break;
        case OP_STOR:
          if (is_code_pointer(c))
            {
              write_log(LOG_DEBUG, "\tSTOR\tR%d\tR%d\t0x%0.8x\n", a, b, c);
              printf("Memory store outside of heap segment at PC: 0x%.8x", (int)PC);
              exit_vm();
            } else
              {
                write_log(LOG_DEBUG, "\tSTOR\tR%d\tR%d\t%d\n", a, b, c);                                
                check_heap_pointer(R[b] + c);
/*                if (R[b] + c >= R[b] + (*(int*)(R[b]-4)))
                  {
                    printf("Memory access outside of bounds for object at 0x%.8x, size %d using offset %d",
                      R[b], (*(int*)(R[b]-4)), c);
                    exit_vm();
                  }*/
                *((int*)(R[b] + c)) =  R[a];
              }
          break;
                                
        case OP_RINT:
          write_log(LOG_DEBUG, "\tRINT\tR%d\n", a);
          scanf("%d", &(R[a]));
          break;
        case OP_PINT:
          write_log(LOG_DEBUG, "\tPINT\tR%d\n", a);
          printf("%d", R[a]);
          break;
        case OP_RCHR:
          write_log(LOG_DEBUG, "\tRCHR\tR%d\n", a);
          char ch;
          scanf("%c\n", &ch);
          R[a] = ch;
          break;
        case OP_PCHR:
          write_log(LOG_DEBUG, "\tPCHR\tR%d\n", a);
          printf("%c", R[a]);
          break;
                        
        case OP_HALT:
          exit_vm();
                        
        default:
          printf("Unknown opcode encountered: %d.\n", opcode);
          exit_vm();
        }
    }
}

/* Print a disassemble of code_mem to stdout. */
void
disassemble(int start, int size)
{
  int pc = start;  
  int a, b, c;
  
  check_code_pointer(pc);
  /* the disassembler loop. */
  while (pc < start + size)
    {
      printf("0x%.8x", pc);
      int op = decode(pc, &a, &b, &c);

      switch (op)
        {
        case OP_LINT:
          printf("\tLINT\tR%d\t", a);
          if (is_code_pointer(c))
            printf("0x%.8x\n", c);
          else
            printf("%d\n", c);
          break;
        case OP_ADD:
          printf("\tADD\tR%d\tR%d\tR%d\n", a, b, c);
          break;
        case OP_SUB:
          printf("\tSUB\tR%d\tR%d\tR%d\n", a, b, c);
          break;
        case OP_MUL:
          printf("\tMUL\tR%d\tR%d\tR%d\n", a, b, c);
          break;
        case OP_DIV:
          printf("\tMOD\tR%d\tR%d\tR%d\n", a, b, c);
          break;
                                
        case OP_ISLT:
          printf("\tISLT\tR%d\tR%d\tR%d\n", a, b, c);
          break;
        case OP_ISLE:
          printf("\tISLE\tR%d\tR%d\tR%d\n", a, b, c);
          break;
        case OP_ISEQ:
          printf("\tISEQ\tR%d\tR%d\tR%d\n", a, b, c);
          break;

        // ...  complter ...
                            
        case OP_ALOC:
          printf("\tALOC\tR%d\tR%d\n", a, b);
          break;
        case OP_LOAD:
          printf("\tLOAD\tR%d\tR%d\t", a, b);
          if (is_code_pointer(c))
            printf("0x%.8x\n", c);
          else
            printf("%d\n", c);                              
          break;
        case OP_STOR:
          printf("\tSTOR\tR%d\tR%d\t", a, b);
          if (is_code_pointer(c))
            printf("0x%.8x\n", c);
          else
            printf("%d\n", c);                              
          break;
                                
        case OP_RINT:
          printf("\tRINT\tR%d\n", a);
          break;
        case OP_PINT:
          printf("\tPINT\tR%d\n", a);
          break;
        case OP_RCHR:
          printf("\tRCHR\tR%d\n", a);
          break;
        case OP_PCHR:
          printf("\tPCHR\tR%d\n", a);
          break;
                        
        case OP_HALT:
          printf("\tHALT\n");
          pc += size;
          break;
                        
        default:
          printf("Unknown opcode encountered: %d.\n", op);
          exit_vm();
        }
        pc += instruction_size(op);
    }
}


/* Dump the code segment to a binary file. */
void
dump_code()
{
  FILE* dump = fopen(CODE_DUMP, "w");
  if (dump != NULL)
    {
      fwrite(code_mem, 1, code_size, dump);
      fclose(dump);
    } else
      perror("Could not dump code memory");
}

/* Return true if the given pointer is in the code segment. */
int
is_code_pointer(int addr)
{
  return (code_mem <= (char*)addr 
          && (char*)addr < (code_mem + code_size));
}

int
is_heap_pointer(int addr)
{
  return (heap_mem <= (char*)addr 
          && (char*)addr < (heap_mem + heap_size));
}

void
check_heap_pointer(int addr)
{
  if (!is_heap_pointer(addr))
    {
      printf("Memory access outside of heap segment: 0x%.8x\n", addr);
      exit_vm();
    }
}

void
check_code_pointer(int addr)
{
  if (!is_code_pointer(addr))
    {
      printf("Memory access outside of code segment: 0x%.8x\n", addr);
      exit_vm();
    }
}



void 
exit_vm()
{
  if (code_mem != NULL)  free(code_mem);
  if (heap_mem != NULL)  free(heap_mem);
  if (inputfile != NULL) fclose(inputfile);
  if (logfile != NULL)   fclose(logfile);
  exit(1);
}
