/*
 * MiniVM project (c) LAMP 2006
 * 
 * Simple parser for the input file.
 * 
 * $Id: parser.c,v 1.1 2006/03/16 15:55:25 dragos Exp $
 */
 
 
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "../main.h"
#include "../emulator/vm.h"
#include "alignloader.h"
#include "../log.h"
#include "labels.h"
#include "tokens.h"
#include "parser.h"

/* The current line number. */
int linenr = 0;

/* The current line in the input file. */
char line[MAX_LINE];
char ident[MAX_LINE];

int c = 0;

/* The address in code where we assemble the next instruction. */
int pc = 0;

/* Lexemes. */
char* identifier();
int   opcode();
int   reg();
int   integer();

/* Non terminals. */
void labeldef();
void instruction();

void skip_whitespace();
void error(const char* str);


/* Read the next line from the input file, and return NULL if
 * end of file is reached.
 */
char *
next_line()
{
  ++linenr;
  c = 0;
  char* str = fgets(line, MAX_LINE, inputfile);
  write_log(LOG_DEBUG, "Read line: '%s'\n", line);
  return str;
}

/* Parse the program from inputfile and store it in code_mem. 
 * To mark the end of the program, a HALT is inserted after the
 * last parsed instruction. */
void
program()
{
  while (next_line() != NULL)
    {
      if (strlen(line) > 0)
        {
          switch (line[0])
            {
            case ';':
              continue;
            case ' ':
            case '\t':
              instruction();
              break;
            default:
              labeldef();
              break;
            }
        }
    }
  /* end marker */
  code_mem[pc] = OP_HALT; 
  end_program = (code_mem + pc);
}

void
labeldef()
{
  identifier();
  skip_whitespace();
  if (line[c++] != ':')
    error("Expected ':'");
  skip_whitespace();
  if (c < (strlen(line) - 1))
    error("Labels should be alone on their line.");

  label_t* label = add_label(ident);
  label->anchored = TRUE;
  label->address  = (int)(code_mem + pc);
  write_log(LOG_DEBUG, "Anchoring label at %0x\n", label->address);
  patch_code(label, pc);
}

/* Handle a reference to a label. If the label has not yet
 * been anchored, or is a completely new label, it is created
 * and the current PC is added to the use-list of the label. 
 * Otherwise, the address of the label is stored at PC. */
void
handle_label()
{
  label_t* label = add_label(identifier());
  if (label->anchored)
    {
      write_log(LOG_DEBUG, "Found anchored label at: %0x\n", label->address);
      PUT_WORD(pc, label->address);
    } else
      {
        write_log(LOG_DEBUG, "Added use site at %0x for label.\n", pc);
        /* add the current pc to the list of patch sites */
        label->uses[label->nuses] = pc;
        ++(label->nuses);
      }       
}

/* Parse an opcode and return it's token. Stop on error. */
int
opcode()
{
  if (identifier() != NULL)
    {
      if (strcasecmp(ident, "LINT") == 0)
        return LINT;
      else if (strcasecmp(ident, "ADD") == 0)
        return ADD;
      else if (strcasecmp(ident, "SUB") == 0)
        return SUB;
      else if (strcasecmp(ident, "MUL") == 0)
        return MUL;
      else if (strcasecmp(ident, "DIV") == 0)
        return DIV;
      else if (strcasecmp(ident, "MOD") == 0)
        return MOD;
      else if (strcasecmp(ident, "ISLT") == 0)
        return ISLT;
      else if (strcasecmp(ident, "ISLE") == 0)
        return ISLE;
      else if (strcasecmp(ident, "ISEQ") == 0)
        return ISEQ;
      // ...  complter ...
      else if (strcasecmp(ident, "ALOC") == 0)
        return ALOC;
      else if (strcasecmp(ident, "LOAD") == 0)
        return LOAD;
      else if (strcasecmp(ident, "STOR") == 0)
        return STOR;
      else if (strcasecmp(ident, "RINT") == 0)
        return RINT;
      else if (strcasecmp(ident, "PINT") == 0)
        return PINT;
      else if (strcasecmp(ident, "RCHR") == 0)
        return RCHR;
      else if (strcasecmp(ident, "PCHR") == 0)
        return PCHR;
      else if (strcasecmp(ident, "HALT") == 0)
        return HALT;
      else
        {
          printf("%s:%d: Unknown opcode '%s'\n", filename, linenr, ident);
          exit_vm();      
          return -1;
        }
    } else
      {
        error("Opcode expected.");
        return -1;
      }
}

int
isIdentTerminator(char c)
{
  return (isblank(c) || c == ':' || c == '\n');
}

int
isIdentStart(char c)
{
  static char* identStart = 
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ|!$%&*+-./:<=>?@^_~";
  return memchr(identStart, c, strlen(identStart)) != NULL;
}

/* Parse and return an identifier. Return NULL on error. */
char *
identifier()
{
  int i = 0;
  skip_whitespace();
  if (c < strlen(line) && isIdentStart(line[c]))
    do 
      {
        ident[i] = line[c];
        ++i;
        ++c;
      } while ((c < strlen(line)) && !isIdentTerminator(line[c]));
        
  ident[i] = '\0';
  if (strlen(ident) > 0)
    return ident;
  else
    return NULL;
}

/* Parse an integer value. Return the parsed value. */
int
integer()
{
  int i = 0;
  skip_whitespace();
  while ((c < strlen(line)) && isdigit(line[c]))
    {
      ident[i] = line[c];
      ++i;
      ++c;
    }
  ident[i] = '\0';
  if (strlen(ident) > 0)
    return atoi(ident);
  else
    {
      error("Expected integer literal.");
      return 0;
    }
}

/* Parse a register in the form [R|r][0..31]*/
int
reg()
{
  skip_whitespace();
  if (c < strlen(line) && (toupper(line[c]) == 'R'))
    {
      ++c;
      int r = integer();
      if (r < REGISTERS)
        return r;
      else
        error("Invalid register.");
    } else
      error("Expected register.");
  return -1;
}

void
skip_whitespace()
{
  while ((c < strlen(line)) && isblank(line[c])) ++c;
}

/* Report an error at the current line number and exit the vm. */
void
error(const char* str)
{
  printf("%s:%d: %s\n", filename, linenr, str);
  exit_vm();
}

